use Test::More tests => 1;

BEGIN { use_ok( 'POE::Component::Server::BigBrother' ); };

diag( "Testing POE::Component::Server::BigBrother-$POE::Component::Server::BigBrother::VERSION, POE-$POE::VERSION, Perl $], $^X" );
